@@ -6,7 +6,7 @@ extern "C" {
 #endif
 
 #include "dict.h"
-#include "pair_list.h"
+#include "hashtable.h"
 #include "state.h"
 
 typedef struct {
@@ -54,17 +54,14 @@ multidict_view_clear(_Multidict_ViewObject *self)
 static inline Py_ssize_t
 multidict_view_len(_Multidict_ViewObject *self)
 {
-    return pair_list_len(&self->md->pairs);
+    return md_len(self->md);
 }
 
 static inline PyObject *
-multidict_view_richcompare(PyObject *self, PyObject *other, int op)
+multidict_view_richcompare(_Multidict_ViewObject *self, PyObject *other, int op)
 {
     int tmp;
-    Py_ssize_t self_size = PyObject_Length(self);
-    if (self_size < 0) {
-        return NULL;
-    }
+    Py_ssize_t self_size = md_len(self->md);
     Py_ssize_t size = PyObject_Length(other);
     if (size < 0) {
         PyErr_Clear();
@@ -76,12 +73,12 @@ multidict_view_richcompare(PyObject *self, PyObject *other, int op)
         case Py_LT:
             if (self_size >= size)
                 Py_RETURN_FALSE;
-            return PyObject_RichCompare(self, other, Py_LE);
+            return multidict_view_richcompare(self, other, Py_LE);
         case Py_LE:
             if (self_size > size) {
                 Py_RETURN_FALSE;
             }
-            iter = PyObject_GetIter(self);
+            iter = PyObject_GetIter((PyObject *)self);
             if (iter == NULL) {
                 goto fail;
             }
@@ -104,16 +101,23 @@ multidict_view_richcompare(PyObject *self, PyObject *other, int op)
         case Py_EQ:
             if (self_size != size)
                 Py_RETURN_FALSE;
-            return PyObject_RichCompare(self, other, Py_LE);
+            return multidict_view_richcompare(self, other, Py_LE);
         case Py_NE:
-            tmp = PyObject_RichCompareBool(self, other, Py_EQ);
-            if (tmp < 0)
+            item = multidict_view_richcompare(self, other, Py_EQ);
+            if (item == NULL) {
                 goto fail;
-            return PyBool_FromLong(!tmp);
+            }
+            if (item == Py_True) {
+                Py_DECREF(item);
+                Py_RETURN_FALSE;
+            } else {
+                Py_DECREF(item);
+                Py_RETURN_TRUE;
+            }
         case Py_GT:
             if (self_size <= size)
                 Py_RETURN_FALSE;
-            return PyObject_RichCompare(self, other, Py_GE);
+            return multidict_view_richcompare(self, other, Py_GE);
         case Py_GE:
             if (self_size < size) {
                 Py_RETURN_FALSE;
@@ -123,7 +127,7 @@ multidict_view_richcompare(PyObject *self, PyObject *other, int op)
                 goto fail;
             }
             while ((item = PyIter_Next(iter))) {
-                tmp = PySequence_Contains(self, item);
+                tmp = PySequence_Contains((PyObject *)self, item);
                 if (tmp < 0) {
                     goto fail;
                 }
@@ -152,7 +156,7 @@ static inline PyObject *
 multidict_itemsview_new(MultiDictObject *md)
 {
     _Multidict_ViewObject *mv = PyObject_GC_New(
-        _Multidict_ViewObject, md->pairs.state->ItemsViewType);
+        _Multidict_ViewObject, md->state->ItemsViewType);
     if (mv == NULL) {
         return NULL;
     }
@@ -184,7 +188,7 @@ multidict_itemsview_repr(_Multidict_ViewObject *self)
         Py_ReprLeave((PyObject *)self);
         return NULL;
     }
-    PyObject *ret = pair_list_repr(&self->md->pairs, name, true, true);
+    PyObject *ret = md_repr(self->md, name, true, true);
     Py_ReprLeave((PyObject *)self);
     Py_CLEAR(name);
     return ret;
@@ -214,7 +218,7 @@ _multidict_itemsview_parse_item(_Multidict_ViewObject *self, PyObject *arg,
         *pvalue = Py_NewRef(PyTuple_GET_ITEM(arg, 1));
     }
 
-    *pidentity = pair_list_calc_identity(&self->md->pairs, key);
+    *pidentity = md_calc_identity(self->md, key);
     Py_DECREF(key);
     if (*pidentity == NULL) {
         if (pkey != NULL) {
@@ -233,7 +237,7 @@ _multidict_itemsview_parse_item(_Multidict_ViewObject *self, PyObject *arg,
     return 1;
 }
 
-static int
+static inline int
 _set_add(PyObject *set, PyObject *key, PyObject * value)
 {
     PyObject *tpl = PyTuple_Pack(2, key, value);
@@ -255,8 +259,7 @@ multidict_itemsview_and1(_Multidict_ViewObject *self, PyObject *other)
     PyObject *value2 = NULL;
     PyObject *arg = NULL;
     PyObject *ret = NULL;
-
-    pair_list_pos_t pos;
+    md_finder_t finder = {0};
 
     PyObject *iter = PyObject_GetIter(other);
     if (iter == NULL) {
@@ -280,29 +283,28 @@ multidict_itemsview_and1(_Multidict_ViewObject *self, PyObject *other)
             continue;
         }
 
-        pair_list_init_pos(&self->md->pairs, &pos);
+        if (md_init_finder(self->md, identity, &finder) < 0) {
+            assert(PyErr_Occurred());
+            goto fail;
+        }
 
-        while (true) {
-            tmp = pair_list_next_by_identity(&self->md->pairs, &pos,
-                                             identity, &key2, &value2);
+        while ((tmp = md_find_next(&finder, &key2, &value2)) > 0) {
+            tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
             if (tmp < 0) {
                 goto fail;
-            } else if (tmp == 0) {
-                break;
-            } else {
-                tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
-                if (tmp < 0) {
+            }
+            if (tmp > 0) {
+                if (_set_add(ret, key2, value2) < 0) {
                     goto fail;
                 }
-                if (tmp > 0) {
-                    if (_set_add(ret, key2, value2) < 0) {
-                        goto fail;
-                    }
-                }
             }
             Py_CLEAR(key2);
             Py_CLEAR(value2);
         }
+        if (tmp < 0) {
+            goto fail;
+        }
+        md_finder_cleanup(&finder);
         Py_CLEAR(arg);
         Py_CLEAR(identity);
         Py_CLEAR(key);
@@ -314,6 +316,7 @@ multidict_itemsview_and1(_Multidict_ViewObject *self, PyObject *other)
     Py_CLEAR(iter);
     return ret;
 fail:
+    md_finder_cleanup(&finder);
     Py_CLEAR(arg);
     Py_CLEAR(identity);
     Py_CLEAR(key);
@@ -334,8 +337,7 @@ multidict_itemsview_and2(_Multidict_ViewObject *self, PyObject *other)
     PyObject *value2 = NULL;
     PyObject *arg = NULL;
     PyObject *ret = NULL;
-
-    pair_list_pos_t pos;
+    md_finder_t finder = {0};
 
     PyObject *iter = PyObject_GetIter(other);
     if (iter == NULL) {
@@ -359,28 +361,27 @@ multidict_itemsview_and2(_Multidict_ViewObject *self, PyObject *other)
             continue;
         }
 
-        pair_list_init_pos(&self->md->pairs, &pos);
+        if (md_init_finder(self->md, identity, &finder) < 0) {
+            assert(PyErr_Occurred());
+            goto fail;
+        }
 
-        while (true) {
-            tmp = pair_list_next_by_identity(&self->md->pairs, &pos,
-                                             identity, NULL, &value2);
+        while ((tmp = md_find_next(&finder, NULL, &value2)) > 0) {
+            tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
             if (tmp < 0) {
                 goto fail;
-            } else if (tmp == 0) {
-                break;
-            } else {
-                tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
-                if (tmp < 0) {
+            }
+            if (tmp > 0) {
+                if (_set_add(ret, key, value2) < 0) {
                     goto fail;
                 }
-                if (tmp > 0) {
-                    if (_set_add(ret, key, value2) < 0) {
-                        goto fail;
-                    }
-                }
             }
             Py_CLEAR(value2);
         }
+        if (tmp < 0) {
+            goto fail;
+        }
+        md_finder_cleanup(&finder);
         Py_CLEAR(arg);
         Py_CLEAR(identity);
         Py_CLEAR(key);
@@ -392,6 +393,7 @@ multidict_itemsview_and2(_Multidict_ViewObject *self, PyObject *other)
     Py_CLEAR(iter);
     return ret;
 fail:
+    md_finder_cleanup(&finder);
     Py_CLEAR(arg);
     Py_CLEAR(identity);
     Py_CLEAR(key);
@@ -435,8 +437,7 @@ multidict_itemsview_or1(_Multidict_ViewObject *self, PyObject *other)
     PyObject *value2 = NULL;
     PyObject *arg = NULL;
     PyObject *ret = NULL;
-
-    pair_list_pos_t pos;
+    md_finder_t finder = {0};
 
     PyObject *iter = PyObject_GetIter(other);
     if (iter == NULL) {
@@ -463,30 +464,30 @@ multidict_itemsview_or1(_Multidict_ViewObject *self, PyObject *other)
             continue;
         }
 
-        pair_list_init_pos(&self->md->pairs, &pos);
+        if (md_init_finder(self->md, identity, &finder) < 0) {
+            assert(PyErr_Occurred());
+            goto fail;
+        }
 
-        while (true) {
-            tmp = pair_list_next_by_identity(&self->md->pairs, &pos,
-                                             identity, NULL, &value2);
+        while ((tmp = md_find_next(&finder, NULL, &value2)) > 0) {
+            tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
             if (tmp < 0) {
                 goto fail;
-            } else if (tmp == 0) {
-                if (PySet_Add(ret, arg) < 0) {
-                    goto fail;
-                }
+            }
+            if (tmp > 0) {
+                Py_CLEAR(value2);
                 break;
-            } else {
-                tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
-                if (tmp < 0) {
-                    goto fail;
-                }
-                if (tmp > 0) {
-                    Py_CLEAR(value2);
-                    break;
-                }
             }
             Py_CLEAR(value2);
         }
+        if (tmp < 0) {
+            goto fail;
+        } else if (tmp == 0) {
+            if (PySet_Add(ret, arg) < 0) {
+                goto fail;
+            }
+        }
+        md_finder_cleanup(&finder);
         Py_CLEAR(arg);
         Py_CLEAR(identity);
         Py_CLEAR(key);
@@ -498,6 +499,7 @@ multidict_itemsview_or1(_Multidict_ViewObject *self, PyObject *other)
     Py_CLEAR(iter);
     return ret;
 fail:
+    md_finder_cleanup(&finder);
     Py_CLEAR(arg);
     Py_CLEAR(identity);
     Py_CLEAR(key);
@@ -518,7 +520,7 @@ multidict_itemsview_or2(_Multidict_ViewObject *self, PyObject *other)
     PyObject *arg = NULL;
     PyObject *tmp_set = NULL;
 
-    pair_list_pos_t pos;
+    md_pos_t pos;
 
     PyObject *ret = PySet_New(other);
     if (ret == NULL) {
@@ -553,11 +555,10 @@ multidict_itemsview_or2(_Multidict_ViewObject *self, PyObject *other)
     }
     Py_CLEAR(iter);
 
-    pair_list_init_pos(&self->md->pairs, &pos);
+    md_init_pos(self->md, &pos);
 
     while (true) {
-        int tmp = pair_list_next(&self->md->pairs, &pos,
-                                 &identity, &key, &value);
+        int tmp = md_next(self->md, &pos, &identity, &key, &value);
         if (tmp < 0) {
             goto fail;
         } else if (tmp == 0) {
@@ -629,7 +630,7 @@ multidict_itemsview_sub1(_Multidict_ViewObject *self, PyObject *other)
     PyObject *ret = NULL;
     PyObject *tmp_set = NULL;
 
-    pair_list_pos_t pos;
+    md_pos_t pos;
 
     PyObject *iter = PyObject_GetIter(other);
     if (iter == NULL) {
@@ -664,11 +665,10 @@ multidict_itemsview_sub1(_Multidict_ViewObject *self, PyObject *other)
     }
     Py_CLEAR(iter);
 
-    pair_list_init_pos(&self->md->pairs, &pos);
+    md_init_pos(self->md, &pos);
 
     while (true) {
-        int tmp = pair_list_next(&self->md->pairs, &pos,
-                                 &identity, &key, &value);
+        int tmp = md_next(self->md, &pos, &identity, &key, &value);
         if (tmp < 0) {
             goto fail;
         } else if (tmp == 0) {
@@ -715,8 +715,7 @@ multidict_itemsview_sub2(_Multidict_ViewObject *self, PyObject *other)
     PyObject *value2 = NULL;
     PyObject *ret = NULL;
     PyObject *iter = PyObject_GetIter(other);
-
-    pair_list_pos_t pos;
+    md_finder_t finder = {0};
 
     if (iter == NULL) {
         if (PyErr_ExceptionMatches(PyExc_TypeError)) {
@@ -742,31 +741,30 @@ multidict_itemsview_sub2(_Multidict_ViewObject *self, PyObject *other)
             continue;
         }
 
-        pair_list_init_pos(&self->md->pairs, &pos);
+        if (md_init_finder(self->md, identity, &finder) < 0) {
+            assert(PyErr_Occurred());
+            goto fail;
+        }
 
-        while (true) {
-            tmp = pair_list_next_by_identity(&self->md->pairs, &pos,
-                                             identity, NULL, &value2);
+        while ((tmp = md_find_next(&finder, NULL, &value2)) > 0) {
+            tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
             if (tmp < 0) {
                 goto fail;
-            } else if (tmp == 0) {
-                if (PySet_Add(ret, arg) < 0) {
-                    goto fail;
-                }
+            }
+            if (tmp > 0) {
+                Py_CLEAR(value2);
                 break;
-            } else {
-                tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
-                if (tmp < 0) {
-                    goto fail;
-                }
-                if (tmp > 0) {
-                    Py_CLEAR(value2);
-                    break;
-                }
             }
             Py_CLEAR(value2);
         }
-
+        if (tmp < 0 ) {
+            goto fail;
+        } else if (tmp == 0) {
+            if (PySet_Add(ret, arg) < 0) {
+                goto fail;
+            }
+        }
+        md_finder_cleanup(&finder);
         Py_CLEAR(arg);
         Py_CLEAR(identity);
         Py_CLEAR(key);
@@ -778,6 +776,7 @@ multidict_itemsview_sub2(_Multidict_ViewObject *self, PyObject *other)
     Py_CLEAR(iter);
     return ret;
 fail:
+    md_finder_cleanup(&finder);
     Py_CLEAR(arg);
     Py_CLEAR(identity);
     Py_CLEAR(key);
@@ -874,64 +873,88 @@ fail:
 static inline int
 multidict_itemsview_contains(_Multidict_ViewObject *self, PyObject *obj)
 {
-    PyObject *akey  = NULL,
-             *aval  = NULL,
-             *bkey  = NULL,
-             *bval  = NULL,
-             *iter  = NULL,
-             *item  = NULL;
-    int ret1, ret2;
+    PyObject *identity = NULL;
+    PyObject *key = NULL;
+    PyObject *value = NULL;
+    PyObject *value2 = NULL;
+    int tmp;
+    int ret = 0;
+    md_finder_t finder = {0};
 
-    if (!PyTuple_Check(obj) || PyTuple_GET_SIZE(obj) != 2) {
-        return 0;
-    }
-
-    bkey = PyTuple_GET_ITEM(obj, 0);
-    bval = PyTuple_GET_ITEM(obj, 1);
-
-    iter = multidict_itemsview_iter(self);
-    if (iter == NULL) {
-        return 0;
-    }
-
-    while ((item = PyIter_Next(iter)) != NULL) {
-        akey = PyTuple_GET_ITEM(item, 0);
-        aval = PyTuple_GET_ITEM(item, 1);
-
-        ret1 = PyObject_RichCompareBool(akey, bkey, Py_EQ);
-        if (ret1 < 0) {
-            Py_DECREF(iter);
-            Py_DECREF(item);
+    if (PyTuple_CheckExact(obj)) {
+        if (PyTuple_GET_SIZE(obj) != 2) {
+            return 0;
+        }
+        key = Py_NewRef(PyTuple_GET_ITEM(obj, 0));
+        value = Py_NewRef(PyTuple_GET_ITEM(obj, 1));
+    } else if (PyList_CheckExact(obj)) {
+        if (PyList_GET_SIZE(obj) != 2) {
+            return 0;
+        }
+        key = Py_NewRef(PyList_GET_ITEM(obj, 0));
+        value = Py_NewRef(PyList_GET_ITEM(obj, 1));
+    } else {
+        tmp = PyObject_Length(obj);
+        if (tmp < 0) {
+            PyErr_Clear();
+            return 0;
+        }
+        if (tmp != 2) {
+            return 0;
+        }
+        key = PySequence_GetItem(obj, 0);
+        if (key == NULL) {
             return -1;
         }
-        ret2 = PyObject_RichCompareBool(aval, bval, Py_EQ);
-        if (ret2 < 0) {
-            Py_DECREF(iter);
-            Py_DECREF(item);
+        value = PySequence_GetItem(obj, 1);
+        if (value == NULL) {
             return -1;
         }
-        if (ret1 > 0 && ret2 > 0)
-        {
-            Py_DECREF(iter);
-            Py_DECREF(item);
-            return 1;
+    }
+
+    identity = md_calc_identity(self->md, key);
+    if (identity == NULL) {
+        PyErr_Clear();
+        ret = 0;
+        goto done;
+    }
+
+    if (md_init_finder(self->md, identity, &finder) < 0) {
+        assert(PyErr_Occurred());
+        ret = -1;
+        goto done;
+    }
+
+    while ((tmp = md_find_next(&finder, NULL, &value2)) > 0) {
+        tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
+        Py_CLEAR(value2);
+        if (tmp < 0) {
+            ret = -1;
+            goto done;
         }
-
-        Py_DECREF(item);
+        if (tmp > 0) {
+            ret = 1;
+            goto done;
+        }
+    }
+    if (tmp < 0) {
+        ret = -1;
+        goto done;
     }
 
-    Py_DECREF(iter);
-
-    if (PyErr_Occurred()) {
-        return -1;
-    }
-
-    return 0;
+ done:
+    md_finder_cleanup(&finder);
+    Py_CLEAR(identity);
+    Py_CLEAR(key);
+    Py_CLEAR(value);
+    ASSERT_CONSISTENT(self->md, false);
+    return ret;
 }
 
 static inline PyObject *
 multidict_itemsview_isdisjoint(_Multidict_ViewObject *self, PyObject *other)
 {
+    md_finder_t finder = {0};
     PyObject *iter = PyObject_GetIter(other);
     if (iter == NULL) {
         return NULL;
@@ -941,8 +964,6 @@ multidict_itemsview_isdisjoint(_Multidict_ViewObject *self, PyObject *other)
     PyObject *value = NULL;
     PyObject *value2 = NULL;
 
-    pair_list_pos_t pos;
-
     while ((arg = PyIter_Next(iter))) {
         int tmp = _multidict_itemsview_parse_item(self, arg,
                                                   &identity, NULL, &value);
@@ -953,31 +974,31 @@ multidict_itemsview_isdisjoint(_Multidict_ViewObject *self, PyObject *other)
             continue;
         }
 
-        pair_list_init_pos(&self->md->pairs, &pos);
+        if (md_init_finder(self->md, identity, &finder) < 0) {
+            assert(PyErr_Occurred());
+            goto fail;
+        }
 
-        while (true) {
-            tmp = pair_list_next_by_identity(&self->md->pairs, &pos,
-                                             identity, NULL, &value2);
+        while ((tmp = md_find_next(&finder, NULL, &value2)) > 0) {
+            tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
+            Py_CLEAR(value2);
             if (tmp < 0) {
                 goto fail;
-            } else if (tmp == 0) {
-                Py_CLEAR(value2);
-                break;
-            } else {
-                tmp = PyObject_RichCompareBool(value, value2, Py_EQ);
-                Py_CLEAR(value2);
-                if (tmp < 0) {
-                    goto fail;
-                }
-                if (tmp > 0) {
-                    Py_CLEAR(iter);
-                    Py_CLEAR(arg);
-                    Py_CLEAR(identity);
-                    Py_CLEAR(value);
-                    Py_RETURN_FALSE;
-                }
+            }
+            if (tmp > 0) {
+                md_finder_cleanup(&finder);
+                Py_CLEAR(iter);
+                Py_CLEAR(arg);
+                Py_CLEAR(identity);
+                Py_CLEAR(value);
+                ASSERT_CONSISTENT(self->md, false);
+                Py_RETURN_FALSE;
             }
         }
+        if (tmp < 0) {
+            goto fail;
+        }
+        md_finder_cleanup(&finder);
         Py_CLEAR(arg);
         Py_CLEAR(identity);
         Py_CLEAR(value);
@@ -986,8 +1007,10 @@ multidict_itemsview_isdisjoint(_Multidict_ViewObject *self, PyObject *other)
     if (PyErr_Occurred()) {
         return NULL;
     }
+    ASSERT_CONSISTENT(self->md, false);
     Py_RETURN_TRUE;
 fail:
+    md_finder_cleanup(&finder);
     Py_CLEAR(iter);
     Py_CLEAR(arg);
     Py_CLEAR(identity);
@@ -1052,7 +1075,7 @@ static inline PyObject *
 multidict_keysview_new(MultiDictObject *md)
 {
     _Multidict_ViewObject *mv = PyObject_GC_New(
-        _Multidict_ViewObject, md->pairs.state->KeysViewType);
+        _Multidict_ViewObject, md->state->KeysViewType);
     if (mv == NULL) {
         return NULL;
     }
@@ -1076,7 +1099,7 @@ multidict_keysview_repr(_Multidict_ViewObject *self)
     if (name == NULL) {
         return NULL;
     }
-    PyObject *ret = pair_list_repr(&self->md->pairs, name, true, false);
+    PyObject *ret = md_repr(self->md, name, true, false);
     Py_CLEAR(name);
     return ret;
 }
@@ -1104,7 +1127,7 @@ multidict_keysview_and1(_Multidict_ViewObject *self, PyObject *other)
             Py_CLEAR(key);
             continue;
         }
-        int tmp = pair_list_contains(&self->md->pairs, key, &key2);
+        int tmp = md_contains(self->md, key, &key2);
         if (tmp < 0) {
             goto fail;
         }
@@ -1151,7 +1174,7 @@ multidict_keysview_and2(_Multidict_ViewObject *self, PyObject *other)
             Py_CLEAR(key);
             continue;
         }
-        int tmp = pair_list_contains(&self->md->pairs, key, NULL);
+        int tmp = md_contains(self->md, key, NULL);
         if (tmp < 0) {
             goto fail;
         }
@@ -1223,7 +1246,7 @@ multidict_keysview_or1(_Multidict_ViewObject *self, PyObject *other)
             Py_CLEAR(key);
             continue;
         }
-        int tmp = pair_list_contains(&self->md->pairs, key, NULL);
+        int tmp = md_contains(self->md, key, NULL);
         if (tmp < 0) {
             goto fail;
         }
@@ -1274,7 +1297,7 @@ multidict_keysview_or2(_Multidict_ViewObject *self, PyObject *other)
             Py_CLEAR(key);
             continue;
         }
-        identity = pair_list_calc_identity(&self->md->pairs, key);
+        identity = md_calc_identity(self->md, key);
         if (identity == NULL) {
             goto fail;
         }
@@ -1289,11 +1312,11 @@ multidict_keysview_or2(_Multidict_ViewObject *self, PyObject *other)
     }
     Py_CLEAR(iter);
 
-    pair_list_pos_t pos;
-    pair_list_init_pos(&self->md->pairs, &pos);
+    md_pos_t pos;
+    md_init_pos(self->md, &pos);
 
     while (true) {
-        int tmp = pair_list_next(&self->md->pairs, &pos, &identity, &key, NULL);
+        int tmp = md_next(self->md, &pos, &identity, &key, NULL);
         if (tmp < 0) {
             goto fail;
         } else if (tmp == 0) {
@@ -1371,7 +1394,7 @@ multidict_keysview_sub1(_Multidict_ViewObject *self, PyObject *other)
             Py_CLEAR(key);
             continue;
         }
-        tmp = pair_list_contains(&self->md->pairs, key, &key2);
+        tmp = md_contains(self->md, key, &key2);
         if (tmp < 0) {
             goto fail;
         }
@@ -1419,7 +1442,7 @@ multidict_keysview_sub2(_Multidict_ViewObject *self, PyObject *other)
             Py_CLEAR(key);
             continue;
         }
-        tmp = pair_list_contains(&self->md->pairs, key, NULL);
+        tmp = md_contains(self->md, key, NULL);
         if (tmp < 0) {
             goto fail;
         }
@@ -1529,7 +1552,7 @@ fail:
 static inline int
 multidict_keysview_contains(_Multidict_ViewObject *self, PyObject *key)
 {
-    return pair_list_contains(&self->md->pairs, key, NULL);
+    return md_contains(self->md, key, NULL);
 }
 
 static inline PyObject *
@@ -1541,7 +1564,7 @@ multidict_keysview_isdisjoint(_Multidict_ViewObject *self, PyObject *other)
     }
     PyObject *key = NULL;
     while ((key = PyIter_Next(iter))) {
-        int tmp = pair_list_contains(&self->md->pairs, key, NULL);
+        int tmp = md_contains(self->md, key, NULL);
         Py_CLEAR(key);
         if (tmp < 0) {
             Py_CLEAR(iter);
@@ -1606,7 +1629,7 @@ static inline PyObject *
 multidict_valuesview_new(MultiDictObject *md)
 {
     _Multidict_ViewObject *mv = PyObject_GC_New(
-        _Multidict_ViewObject, md->pairs.state->ValuesViewType);
+        _Multidict_ViewObject, md->state->ValuesViewType);
     if (mv == NULL) {
         return NULL;
     }
@@ -1638,7 +1661,7 @@ multidict_valuesview_repr(_Multidict_ViewObject *self)
         Py_ReprLeave((PyObject *)self);
         return NULL;
     }
-    PyObject *ret = pair_list_repr(&self->md->pairs, name, false, true);
+    PyObject *ret = md_repr(self->md, name, false, true);
     Py_ReprLeave((PyObject *)self);
     Py_CLEAR(name);
     return ret;
